"""Object that deals with the open loop optimal control."""

import os
import shutil
import tensorflow as tf
from typing import Callable, List

from c3.optimizers.optimizer import Optimizer
from c3.utils.utils import log_setup

from c3.libraries.algorithms import algorithms
from c3.libraries.fidelities import fidelities


class OptimalControl(Optimizer):
    """
    Object that deals with the open loop optimal control.

    Parameters
    ----------
    dir_path : str
        Filepath to save results
    fid_func : callable
        infidelity function to be minimized
    fid_subspace : list
        Indeces identifying the subspace to be compared
    pmap : ParameterMap
        Identifiers for the parameter vector
    callback_fids : list of callable
        Additional fidelity function to be evaluated and stored for reference
    algorithm : callable
        From the algorithm library
        Save plots of control signals
    store_unitaries : boolean
        Store propagators as text and pickle
    options : dict
        Options to be passed to the algorithm
    run_name : str
        User specified name for the run, will be used as root folder
    fid_func_kwargs: dict
        Additional kwargs to be passed to the main fidelity function.
    """

    def __init__(
        self,
        fid_func,
        fid_subspace,
        pmap,
        dir_path=None,
        callback_fids=None,
        algorithm=None,
        initial_point: str = "",
        store_unitaries=False,
        options={},
        run_name=None,
        interactive=True,
        include_model=False,
        logger=None,
        fid_func_kwargs={},
        ode_solver=None,
        ode_step_function="schrodinger",
        only_final_state=False,
    ) -> None:
        if type(algorithm) is str:
            algorithm = algorithms[algorithm]
        super().__init__(
            pmap=pmap,
            algorithm=algorithm,
            initial_point=initial_point,
            store_unitaries=store_unitaries,
            logger=logger,
        )
        self.set_fid_func(fid_func)
        self.callback_fids: List[Callable] = []
        if callback_fids:
            self.set_callback_fids(callback_fids)
        self.fid_subspace = fid_subspace
        self.options = options
        self.__dir_path = dir_path
        self.__run_name = run_name
        self.interactive = interactive
        self.update_model = include_model
        self.fid_func_kwargs = fid_func_kwargs
        self.run = (
            self.optimize_controls
        )  # Alias the legacy name for the method running the
        # optimization
        self.set_goal_function(
            ode_solver=ode_solver,
            ode_step_function=ode_step_function,
            only_final_state=only_final_state,
        )

    def set_goal_function(
        self,
        ode_solver=None,
        ode_step_function="schrodinger",
        only_final_state=False,
    ):
        self.ode_solver = ode_solver
        self.ode_step_function = ode_step_function
        self.only_final_state = only_final_state

        if self.ode_solver is not None:
            if self.only_final_state:
                self.goal_function = self.goal_run_ode_only_final
            else:
                self.goal_function = self.goal_run_ode
        else:
            self.goal_function = self.goal_run

    def set_fid_func(self, fid_func) -> None:
        if type(fid_func) is str:
            if self.pmap.model.lindbladian:
                fid = "lindbladian_" + fid_func
            else:
                fid = fid_func
            try:
                self.fid_func = fidelities[fid]
            except KeyError:
                raise Exception(f"C3:ERROR:Unkown goal function: {fid} ")
            print(f"C3:STATUS:Found {fid} in libraries.")
        else:
            self.fid_func = fid_func

    def set_callback_fids(self, callback_fids) -> None:
        if self.pmap.model.lindbladian:
            cb_fids = ["lindbladian_" + f for f in callback_fids]
        else:
            cb_fids = callback_fids
        for cb_fid in cb_fids:
            try:
                cb_fid_func = fidelities[cb_fid]
            except KeyError:
                raise Exception(f"C3:ERROR:Unkown goal function: {cb_fid}")
            print(f"C3:STATUS:Found {cb_fid} in libraries.")
            self.callback_fids.append(cb_fid_func)

    def log_setup(self) -> None:
        """
        Create the folders to store data.
        """
        run_name = self.__run_name
        if run_name is None:
            run_name = "c1_" + self.fid_func.__name__ + "_" + self.algorithm.__name__
        self.logdir = log_setup(self.__dir_path, run_name)
        self.logname = "open_loop.c3log"
        if isinstance(self.exp.created_by, str):
            shutil.copy2(self.exp.created_by, self.logdir)
        if isinstance(self.created_by, str):
            shutil.copy2(self.created_by, self.logdir)

    def load_model_parameters(self, adjust_exp: str) -> None:
        self.pmap.load_values(adjust_exp)
        self.pmap.model.update_model()
        shutil.copy(adjust_exp, os.path.join(self.logdir, "adjust_exp.c3log"))

    def optimize_controls(self, setup_log: bool = True) -> None:
        """
        Apply a search algorithm to your gateset given a fidelity function.
        """
        if setup_log:
            self.log_setup()
        self.start_log()
        self.exp.set_enable_store_unitaries(self.store_unitaries, self.logdir)
        print(f"C3:STATUS:Saving as: {os.path.abspath(self.logdir + self.logname)}")
        index = []
        for name in self.fid_subspace:
            index.append(self.pmap.model.names.index(name))
        self.index = index
        x_init = self.pmap.get_parameters_scaled()
        try:
            self.algorithm(
                x_init,
                fun=self.fct_to_min,
                fun_grad=self.fct_to_min_autograd,
                grad_lookup=self.lookup_gradient,
                options=self.options,
            )
        except KeyboardInterrupt:
            pass
        self.load_best(self.logdir + "best_point_" + self.logname)
        self.end_log()

    @tf.function
    def goal_run(self, current_params: tf.Tensor) -> tf.float64:
        """
        Evaluate the goal function for current parameters.

        Parameters
        ----------
        current_params : tf.Tensor
            Vector representing the current parameter values.

        Returns
        -------
        tf.float64
            Value of the goal function
        """
        self.pmap.set_parameters_scaled(current_params)
        dims = self.pmap.model.dims
        propagators = self.exp.compute_propagators()

        goal = self.fid_func(
            propagators=propagators,
            instructions=self.pmap.instructions,
            index=self.index,
            dims=dims,
            n_eval=self.evaluation + 1,
            **self.fid_func_kwargs,
        )
        self.evaluation += 1
        return goal

    @tf.function
    def goal_run_ode(self, current_params: tf.Tensor) -> tf.float64:
        """
        Evaluate the goal function using ode solver for current parameters.

        Parameters
        ----------
        current_params : tf.Tensor
            Vector representing the current parameter values.

        Returns
        -------
        tf.float64
            Value of the goal function
        """
        self.pmap.set_parameters_scaled(current_params)
        dims = self.pmap.model.dims
        result = self.exp.compute_states(
            solver=self.ode_solver, step_function=self.ode_step_function
        )
        states = result["states"]

        goal = self.fid_func(
            states=states,
            index=self.index,
            dims=dims,
            n_eval=self.evaluation + 1,
            **self.fid_func_kwargs,
        )
        self.evaluation += 1
        return goal

    @tf.function
    def goal_run_ode_only_final(self, current_params: tf.Tensor) -> tf.float64:
        """
        Evaluate the goal function using ode solver for current parameters.

        Parameters
        ----------
        current_params : tf.Tensor
            Vector representing the current parameter values.

        Returns
        -------
        tf.float64
            Value of the goal function
        """
        self.pmap.set_parameters_scaled(current_params)
        dims = self.pmap.model.dims
        result = self.exp.compute_final_state(
            solver=self.ode_solver, step_function=self.ode_step_function
        )
        state = result["states"]

        goal = self.fid_func(
            states=state,
            index=self.index,
            dims=dims,
            n_eval=self.evaluation + 1,
            **self.fid_func_kwargs,
        )
        self.evaluation += 1
        return goal



"""
integration testing module for C1 optimization through two-qubits example
"""

import pickle
import numpy as np
import pytest
import tensorflow as tf

from numpy.testing import assert_array_almost_equal as almost_equal
import c3.libraries.algorithms as algorithms
from c3.libraries.fidelities import state_transfer_from_states

# Libs and helpers
from c3.libraries.propagation import rk4_unitary

with open("test/two_qubit_data.pickle", "rb") as filename:
    test_data = pickle.load(filename)


@pytest.mark.tensorflow
@pytest.mark.unit
def test_signals(get_two_qubit_chip) -> None:
    """Compare generated control signals to stored examples."""
    exp = get_two_qubit_chip
    pmap = exp.pmap
    gen_signal = pmap.generator.generate_signals(pmap.instructions["rx90p[0]"])
    ts = gen_signal["d1"]["ts"]
    np.testing.assert_allclose(ts, test_data["ts"])
    np.testing.assert_allclose(
        actual=gen_signal["d1"]["values"].numpy(),
        desired=test_data["signal"]["d1"]["values"].numpy(),
    )


@pytest.mark.unit
def test_hamiltonians(get_two_qubit_chip) -> None:
    """Compare Hamilonians"""
    model = get_two_qubit_chip.pmap.model
    hdrift, hks = model.get_Hamiltonians()
    assert (hdrift.numpy() - test_data["hdrift"].numpy() < 1).any()
    for key in hks:
        almost_equal(hks[key], test_data["hks"][key])


@pytest.mark.tensorflow
@pytest.mark.integration
def test_propagation(get_two_qubit_chip) -> None:
    """Test that result of the propagation code does not change."""
    GATE_STR = "rx90p[0]"
    exp = get_two_qubit_chip
    pmap = exp.pmap
    instr = pmap.instructions[GATE_STR]
    steps = int((instr.t_end - instr.t_start) * exp.sim_res)
    result = exp.propagation(
        pmap.model,
        pmap.generator,
        pmap.instructions["rx90p[0]"],
        exp.folding_stack[steps],
    )
    propagator = result["U"]
    almost_equal(propagator, test_data["propagator"])


@pytest.mark.unit
def test_init_point(get_OC_optimizer) -> None:
    """Check that a previous best point can be loaded as an initial point."""
    opt = get_OC_optimizer
    opt.load_best("test/best_point_open_loop.c3log")


@pytest.mark.slow
@pytest.mark.tensorflow
@pytest.mark.optimizers
@pytest.mark.integration
def test_optim(get_OC_optimizer) -> None:
    """
    check if optimization result is below 1e-1
    """
    opt = get_OC_optimizer
    assert opt.evaluation == 0
    opt.optimize_controls()
    assert opt.current_best_goal < 0.1
    maxiterKey = "maxiters" if opt.algorithm == algorithms.tf_sgd else "maxiter"
    assert opt.evaluation == opt.options[maxiterKey] - 1


@pytest.mark.slow
@pytest.mark.tensorflow
@pytest.mark.optimizers
@pytest.mark.integration
def test_optim_ode_solver(get_OC_optimizer) -> None:
    """
    check if optimization result is below 1e-1
    """
    opt = get_OC_optimizer
    opt.fid_func = state_transfer_from_states
    opt.set_goal_function(ode_solver="rk4", ode_step_function="schrodinger")

    model = opt.exp.pmap.model
    psi_init = [[0] * model.tot_dim]
    psi_init[0][0] = 1 / np.sqrt(2)
    index = model.get_state_indeces([(1, 0)])[0]
    psi_init[0][index] = 1 / np.sqrt(2)
    target_state = tf.transpose(tf.constant(psi_init, tf.complex128))
    params = {"target": target_state}
    opt.fid_func_kwargs = {"params": params}

    assert opt.evaluation == 0
    opt.optimize_controls()

    if opt.algorithm == algorithms.single_eval:
        assert opt.current_best_goal < 0.5  # TODO - Change this to 0.1
    else:
        assert opt.current_best_goal < 0.3  # TODO - Change this to 0.1

    maxiterKey = "maxiters" if opt.algorithm == algorithms.tf_sgd else "maxiter"
    assert opt.evaluation == opt.options[maxiterKey] - 1


@pytest.mark.slow
@pytest.mark.tensorflow
@pytest.mark.optimizers
@pytest.mark.integration
def test_optim_ode_solver_final(get_OC_optimizer) -> None:
    """
    check if optimization result is below 1e-1
    """
    opt = get_OC_optimizer
    opt.fid_func = state_transfer_from_states
    opt.set_goal_function(
        ode_solver="rk4", ode_step_function="schrodinger", only_final_state=True
    )

    model = opt.exp.pmap.model
    psi_init = [[0] * model.tot_dim]
    psi_init[0][0] = 1 / np.sqrt(2)
    index = model.get_state_indeces([(1, 0)])[0]
    psi_init[0][index] = 1 / np.sqrt(2)
    target_state = tf.transpose(tf.constant(psi_init, tf.complex128))
    params = {"target": target_state}
    opt.fid_func_kwargs = {"params": params}

    assert opt.evaluation == 0
    opt.optimize_controls()
    if opt.algorithm == algorithms.single_eval:
        assert opt.current_best_goal < 0.5  # TODO - Change this to 0.1
    else:
        assert opt.current_best_goal < 0.3  # TODO - Change this to 0.1

    maxiterKey = "maxiters" if opt.algorithm == algorithms.tf_sgd else "maxiter"
    assert opt.evaluation == opt.options[maxiterKey] - 1


@pytest.mark.tensorflow
@pytest.mark.integration
def test_rk4_unitary(get_two_qubit_chip) -> None:
    """Testing that RK4 exists and runs."""
    exp = get_two_qubit_chip
    pmap = exp.pmap
    exp.set_prop_method(rk4_unitary)
    exp.propagation(pmap.model, pmap.generator, pmap.instructions["rx90p[0]"])


@pytest.mark.tensorflow
@pytest.mark.integration
def test_ode_solver(get_two_qubit_chip) -> None:
    """Testing that ODE solver exists and runs for solvers rk4, rk5, tsit5."""
    exp = get_two_qubit_chip
    exp.set_opt_gates(["rx90p[0]"])
    exp.compute_states(solver="rk4")
    exp.compute_states(solver="rk5")
    exp.compute_states(solver="tsit5")
    exp.compute_states(solver="rk4", step_function="von_neumann")
    exp.compute_states(solver="rk5", step_function="von_neumann")
    exp.compute_states(solver="tsit5", step_function="von_neumann")


@pytest.mark.tensorflow
@pytest.mark.integration
def test_compute_final_state(get_two_qubit_chip) -> None:
    """Testing that compute_final_state exists and runs for solvers rk4, rk5, tsit5."""
    exp = get_two_qubit_chip
    exp.set_opt_gates(["rx90p[0]"])
    exp.compute_final_state(solver="rk4")
    exp.compute_final_state(solver="rk5")
    exp.compute_final_state(solver="tsit5")
    exp.compute_final_state(solver="rk4", step_function="von_neumann")
    exp.compute_final_state(solver="rk5", step_function="von_neumann")
    exp.compute_final_state(solver="tsit5", step_function="von_neumann")


@pytest.mark.slow
@pytest.mark.tensorflow
@pytest.mark.integration
def test_lindblad_propagation(get_two_qubit_chip) -> None:
    """Test that result of the propagation code does not change."""
    GATE_STR = "rx90p[0]"
    exp = get_two_qubit_chip
    exp.propagate_batch_size = 360
    pmap = exp.pmap
    model = pmap.model
    model.set_lindbladian(True)
    instr = pmap.instructions[GATE_STR]
    steps = int((instr.t_end - instr.t_start) * exp.sim_res)
    result = exp.propagation(
        model,
        pmap.generator,
        pmap.instructions["rx90p[0]"],
        exp.folding_stack[steps],
    )
    propagator = result["U"]
    almost_equal(propagator, test_data["lindblad_propagator"])


@pytest.mark.tensorflow
@pytest.mark.integration
def test_ode_solver_lindblad(get_two_qubit_chip) -> None:
    """Testing that ODE solver for lindbladian exists and runs for solvers."""
    exp = get_two_qubit_chip
    model = exp.pmap.model
    model.set_lindbladian(True)
    exp.set_opt_gates(["rx90p[0]"])
    exp.compute_states(solver="rk4")
    exp.compute_final_state(solver="rk4")


@pytest.mark.tensorflow
@pytest.mark.integration
def test_state_norm(get_two_qubit_chip) -> None:
    """Testing that the norm of the states is correct for ODE solvers."""
    exp = get_two_qubit_chip
    model = exp.pmap.model
    exp.set_opt_gates(["rx90p[0]"])

    model.set_lindbladian(False)
    output = exp.compute_states(solver="rk4")
    state = output["states"]
    almost_equal(
        tf.linalg.norm(state[-1]), 1, decimal=2
    )  # TODO - Change this to 6 decimal places

    model.set_lindbladian(False)
    output = exp.compute_states(solver="rk4", step_function="von_neumann")
    state = output["states"]
    almost_equal(tf.linalg.trace(state[-1]), 1, decimal=6)

    model.set_lindbladian(True)
    output = exp.compute_states(solver="rk4")
    state = output["states"]
    almost_equal(tf.linalg.trace(state[-1]), 1, decimal=6)
